import tensorflow as tf
from tensorflow import keras
import numpy as np


class SVD(keras.Model):
    """ **SVD** - A Matrix Factorization Algorithm for Collaborative Filtering.

    As demonstrated by Simon Funk in his
    blogpost https://sifter.org/~simon/journal/20061211.html.

    ...

    Attributes
    -----------
    n_users : int
        Total number of users in the dataset.
    n_items : int
        Total number of items in the dataset.
    global_mean : float
        Mean of ratings in the training set.
    embedding_dim : int, optional
        Number of factors for user and item (default is `50`).
    biased : boolean, optional
        Whether bias should be used or not (default is `True`).
    init_mean : float, optional
        Mean of random initilization for embeddings (default is `0`).
    init_std_dev : float, optional
        Standard deviation of random initilization for embeddings (default is `0.1`).
    reg_all : float, optional
        L2 regularization factor for all trainable variables (default is `0.0001`).
    reg_user_embed : float, optional
        L2 regularization factor for user embedding (default is `reg_all`).
    reg_item_embed : float, optional
        L2 regularization factor for item embedding (default is `reg_all`).
    reg_user_bias : float, optional
        L2 regularization factor for user bias (default is `reg_all`).
    reg_item_bias : float, optional
        L2 regularization factor for item bias (default is `reg_all`).
    random_state : integer, optional
        Seed variable, useful for reproducing results (default is `None`).
    """

    def __init__(self, n_users, n_items, global_mean, embedding_dim=50, biased=True,
                 init_mean=0, init_std_dev=0.1, reg_all=0.0001, reg_user_embed=None,
                 reg_item_embed=None, reg_user_bias=None, reg_item_bias=None, random_state=None, **kwargs):
        """
        Parameters
        -----------
        n_users : int
            Total number of users in the dataset.
        n_items : int
            Total number of items in the dataset.
        global_mean : float
            Mean of ratings in the training set.
        embedding_dim : int, optional
            Number of factors for user and item (default is `50`).
        biased : boolean, optional
            Whether bias should be used or not (default is `True`).
        init_mean : float, optional
            Mean of random initilization for embeddings (default is `0`).
        init_std_dev : float, optional
            Standard deviation of random initilization for embeddings (default is `0.1`).
        reg_all : float, optional
            L2 regularization factor for all trainable variables (default is `0.0001`).
        reg_user_embed : float, optional
            L2 regularization factor for user embedding (default is `reg_all`).
        reg_item_embed : float, optional
            L2 regularization factor for item embedding (default is `reg_all`).
        reg_user_bias : float, optional
            L2 regularization factor for user bias (default is `reg_all`).
        reg_item_bias : float, optional
            L2 regularization factor for item bias (default is `reg_all`).
        random_state : integer, optional
            Seed variable, useful for reproducing results (default is `None`).
        """

        super().__init__(**kwargs)
        self.n_users = n_users
        self.n_items = n_items
        self.global_mean = np.float32(global_mean)
        self.embedding_dim = embedding_dim
        self.biased = biased
        self.init_mean = init_mean
        self.init_std_dev = init_std_dev
        self.reg_all = reg_all
        self.reg_user_embed = reg_all if not reg_user_embed else reg_user_embed
        self.reg_item_embed = reg_all if not reg_item_embed else reg_item_embed
        self.reg_user_bias = reg_all if not reg_user_bias else reg_user_bias
        self.reg_item_bias = reg_all if not reg_item_bias else reg_item_bias
        self.random_state = random_state

        self.user_embedding = keras.layers.Embedding(
            input_dim=self.n_users, output_dim=self.embedding_dim,
            embeddings_initializer=tf.keras.initializers.RandomNormal(
                mean=self.init_mean, stddev=self.init_std_dev, seed=self.random_state),
            embeddings_regularizer=tf.keras.regularizers.L2(
                self.reg_user_embed)
        )
        self.item_embedding = keras.layers.Embedding(
            input_dim=self.n_items, output_dim=self.embedding_dim,
            embeddings_initializer=tf.keras.initializers.RandomNormal(
                mean=self.init_mean, stddev=self.init_std_dev, seed=self.random_state),
            embeddings_regularizer=tf.keras.regularizers.L2(
                self.reg_item_embed)
        )
        if self.biased:
            self.user_bias = keras.layers.Embedding(
                input_dim=self.n_users, output_dim=1,
                embeddings_initializer=tf.keras.initializers.Zeros(),
                embeddings_regularizer=tf.keras.regularizers.L2(
                    self.reg_user_bias)
            )
            self.item_bias = keras.layers.Embedding(
                input_dim=self.n_items, output_dim=1,
                embeddings_initializer=tf.keras.initializers.Zeros(),
                embeddings_regularizer=tf.keras.regularizers.L2(
                    self.reg_item_bias)
            )

    def call(self, inputs):
        """Forward pass of input batch."""
        # Separate Inputs
        user, item = inputs[:, 0], inputs[:, 1]

        # Dot Product
        user_embed, item_embed = self.user_embedding(
            user), self.item_embedding(item)
        rating = tf.math.reduce_sum(tf.multiply(
            user_embed, item_embed), 1, keepdims=True)

        # Add global mean and bias if self.bias = True
        if self.biased:
            user_bias, item_bias = self.user_bias(user), self.item_bias(item)
            total_bias = tf.math.add(
                self.global_mean, tf.math.add(user_bias, item_bias))
            rating = tf.math.add(rating, total_bias)

        return rating



import tensorflow as tf
from tensorflow import keras
import numpy as np


class SVDpp(keras.Model):
    """ **SVD++** - An extension of the SVD Model employing implicit feedback.

    As originally demonstrated in the paper -
    Factorization meets the neighborhood: a multifaceted collaborative filtering model
    https://dl.acm.org/doi/10.1145/1401890.1401944 Section 4.

    ...

    Attributes
    -----------
    n_users : int
        Total number of users in the dataset.
    n_items : int
        Total number of items in the dataset.
    global_mean : float
        Mean of ratings in the training set.
    embedding_dim : int, optional
        Number of factors for user and item (default is `50`).
    init_mean : float, optional
        Mean of random initilization for embeddings (default is `0`).
    init_std_dev : float, optional
        Standard deviation of random initilization for embeddings (default is `0.1`).
    reg_all : float, optional
        L2 regularization factor for all trainable variables (default is `0.0001`).
    reg_user_embed : float, optional
        L2 regularization factor for user embedding (default is `reg_all`).
    reg_item_embed : float, optional
        L2 regularization factor for item embedding (default is `reg_all`).
    reg_user_bias : float, optional
        L2 regularization factor for user bias (default is `reg_all`).
    reg_item_bias : float, optional
        L2 regularization factor for item bias (default is `reg_all`).
    random_state : integer, optional
        Seed variable, useful for reproducing results (default is `None`).

    Raises
    ------
    AttributeError
        If the method `implicit_feedback(X)` is not called before calling `fit()`.
    """

    def __init__(self, n_users, n_items, global_mean, embedding_dim=50, init_mean=0, init_std_dev=0.1, reg_all=0.0001,
                 reg_user_embed=None, reg_item_embed=None, reg_impl_embed=None, reg_user_bias=None, reg_item_bias=None, random_state=None, **kwargs):

        super().__init__(**kwargs)
        self.n_users = n_users
        self.n_items = n_items
        self.global_mean = np.float32(global_mean)
        self.embedding_dim = embedding_dim
        self.init_mean = init_mean
        self.init_std_dev = init_std_dev
        self.reg_all = reg_all
        self.reg_user_embed = reg_all if not reg_user_embed else reg_user_embed
        self.reg_item_embed = reg_all if not reg_item_embed else reg_item_embed
        self.reg_impl_embed = reg_all if not reg_impl_embed else reg_impl_embed
        self.reg_user_bias = reg_all if not reg_user_bias else reg_user_bias
        self.reg_item_bias = reg_all if not reg_item_bias else reg_item_bias
        self.random_state = random_state

        self.user_embedding = keras.layers.Embedding(
            input_dim=self.n_users, output_dim=self.embedding_dim,
            embeddings_initializer=tf.keras.initializers.RandomNormal(
                mean=self.init_mean, stddev=self.init_std_dev, seed=self.random_state),
            embeddings_regularizer=tf.keras.regularizers.L2(
                self.reg_user_embed)
        )
        self.item_embedding = keras.layers.Embedding(
            input_dim=self.n_items, output_dim=self.embedding_dim,
            embeddings_initializer=tf.keras.initializers.RandomNormal(
                mean=self.init_mean, stddev=self.init_std_dev, seed=self.random_state),
            embeddings_regularizer=tf.keras.regularizers.L2(
                self.reg_item_embed)
        )
        self.item_implicit_embedding = keras.layers.Embedding(
            input_dim=self.n_items, output_dim=self.embedding_dim,
            embeddings_initializer=tf.keras.initializers.RandomNormal(
                mean=self.init_mean, stddev=self.init_std_dev, seed=self.random_state),
            embeddings_regularizer=tf.keras.regularizers.L2(
                self.reg_impl_embed)
        )
        self.user_bias = keras.layers.Embedding(
            input_dim=self.n_users, output_dim=1,
            embeddings_initializer=tf.keras.initializers.Zeros(),
            embeddings_regularizer=tf.keras.regularizers.L2(self.reg_user_bias)
        )
        self.item_bias = keras.layers.Embedding(
            input_dim=self.n_items, output_dim=1,
            embeddings_initializer=tf.keras.initializers.Zeros(),
            embeddings_regularizer=tf.keras.regularizers.L2(self.reg_item_bias)
        )

    def implicit_feedback(self, X):
        """Maps a user to rated items for implicit feedback.

        Needs to be called before fitting the Model.

        Parameters
        ----------
        X : numpy.ndarray
            User Item table.

        Raises
        ------
        AttributeError
            If this is not called before calling `fit()`.
        """

        self.user_rated_items = [[] for _ in range(self.n_users)]
        for u, i in zip(X[:, 0], X[:, 1]):
            self.user_rated_items[u].append(i)

        # Converts to ragged tensor to be used during forward pass
        self.user_rated_items = tf.ragged.constant(
            self.user_rated_items, dtype=tf.int32)

    def call(self, inputs):
        """Forward pass of input batch."""
        # Separate Inputs
        user, item = inputs[:, 0], inputs[:, 1]

        # Get Embeddings
        user_embed, item_embed = self.user_embedding(
            user), self.item_embedding(item)
        user_bias, item_bias = self.user_bias(user), self.item_bias(item)

        # Gather Rated Items & their lengths
        rated_items = tf.gather(self.user_rated_items, user)
        item_lengths = tf.cast(
            tf.map_fn(tf.shape, rated_items).to_tensor(), dtype=tf.float32)

        # Calculate Implicit Feedback
        implicit_embed = self.item_implicit_embedding(rated_items)
        implicit_embed_sum = tf.reduce_sum(implicit_embed, axis=1)
        moderated_implicit_embed = tf.math.divide(
            implicit_embed_sum, tf.math.sqrt(item_lengths))

        # Calculate Final Rating with bias
        total_user_embed = tf.math.add(user_embed, moderated_implicit_embed)
        rating = tf.math.reduce_sum(tf.multiply(
            total_user_embed, item_embed), 1, keepdims=True)
        total_bias = tf.math.add(
            self.global_mean, tf.math.add(user_bias, item_bias))
        rating = tf.math.add(rating, total_bias)

        return rating



import unittest
import numpy as np
from tfrec.datasets import fetch_ml_100k
from tfrec.utils.model_utils import preprocess_and_split


class TestPreprocess(unittest.TestCase):

    def setUp(self):
        self.dataset = fetch_ml_100k()

    def test_preprocess_instance(self):
        preprocessed_dataset, user_item_encodings = preprocess_and_split(self.dataset)

        (x_train, y_train), (x_test, y_test) = preprocessed_dataset

        self.assertIsInstance(x_train, np.ndarray)
        self.assertIsInstance(y_train, np.ndarray)
        self.assertIsInstance(x_test, np.ndarray)
        self.assertIsInstance(y_test, np.ndarray)

        self.assertEqual(len(x_train) + len(x_test), len(self.dataset))

        (user_to_encoded, encoded_to_user, item_to_encoded, encoded_to_item) = user_item_encodings

        for instance in user_item_encodings:
            self.assertIsInstance(instance, dict)

        self.assertEqual(len(user_to_encoded), len(encoded_to_user))
        self.assertEqual(len(item_to_encoded), len(encoded_to_item))



import unittest
import pandas as pd
from tfrec.datasets import fetch_ml_100k


class Test_ml_100k(unittest.TestCase):
    def test_instance(self):
        dataset = fetch_ml_100k()
        self.assertIsInstance(dataset, pd.core.frame.DataFrame)



import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split, KFold


def preprocess_and_split(df, shuffle=True, split=True, test_size=0.2, random_state=None):
    """Preprocesses and Splits the dataset if necessary.

    **Encodes** & **Decodes** User and Item, each from `0 to len(user) and len(item)`.

    Splits into **Train** and **Test** and shuffles(if `shuffle = True`).

    Returns Dataset as numpy arrays and **Encoding** and **Decodings** of users and items.

    Parameters
    ----------
    df : pandas.core.frame.DataFrame
        Dataframe of shape `(m,3)` with columns in the order - `user`, `item`, `rating`.
    shuffle : boolean, optional
        Boolean to determine should the dataset be shuffled after encoding (default is `True`).
    split : boolean, optional
        Boolean to determine if the dataset should be split into train and test set (default is `True`).
    test_size : float, optional
        Size of test split (default is `0.2`).
    random_state : integer, optional
        Seed variable, useful for reproducing results (default is `None`).

    Returns
    -------
    dataset, user_item_encodings : tuple
        Dataset contains tuple of train and test variables if `shuffle = True`.

        **dataset**

            (x_train, y_train), (x_test, y_test).

        **user_item_encodings**

            (user_to_encoded, encoded_to_user,item_to_encoded, encoded_to_item).
    """

    # Get unique ids
    user_ids = df.iloc[:, 0].unique().tolist()
    item_ids = df.iloc[:, 1].unique().tolist()

    # Create encoding and decoding maps for users
    user_to_encoded = {user_id: encoded_id for encoded_id,
                       user_id in enumerate(user_ids)}
    encoded_to_user = {encoded_id: user_id for encoded_id,
                       user_id in enumerate(user_ids)}

    # Create encoding and decoding maps for items
    item_to_encoded = {item_id: encoded_id for encoded_id,
                       item_id in enumerate(item_ids)}
    encoded_to_item = {encoded_id: item_id for encoded_id,
                       item_id in enumerate(item_ids)}

    user_item_encodings = (user_to_encoded, encoded_to_user,
                           item_to_encoded, encoded_to_item)

    # Encode Dataset
    df_encoded = pd.DataFrame([], columns=['user', 'item', 'rating'])
    df_encoded['user'] = df.iloc[:, 0].map(user_to_encoded)
    df_encoded['item'] = df.iloc[:, 1].map(item_to_encoded)
    df_encoded['rating'] = df.iloc[:, 2].values.astype(np.float32)

    # Shuffle if shuffle = True
    if shuffle:
        df_encoded = df_encoded.sample(frac=1, random_state=None)

    # Dataframe -> Numpy Arrays
    x = df_encoded[['user', 'item']].values
    y = df_encoded[['rating']].values
    dataset = x, y

    # Split if split = True
    if split:
        x_train, x_test, y_train, y_test = train_test_split(
            x, y, test_size=test_size, random_state=random_state
        )
        dataset = (x_train, y_train), (x_test, y_test)

    return dataset, user_item_encodings


def cross_validate(model, X, y, folds=5, epochs=5, batch_size=32, callbacks=None, shuffle=False, random_state=None):
    """Performs K-Fold cross validation on data.

    Parameters
    ----------
    model : tf.keras.Model
        A compiled model with appropriate metrics to keep track of.
    X : numpy.ndarray
        Numpy array of shape `(m, 2)` of user-item pairs.
    y : numpy.ndarray
        Numpy array of shape `(m,)` of target rating.
    folds : integer, optional
        Number of splits for K Fold (default is `5`).
    epochs : integer, optional
        Epochs to fit the model (default is `10`).
    batch_size : integer, optional
        Batch_size to be passed in while fitting (default is `32`).
    callbacks : list, optional
        List of callbacks to be passed in while fitting (default `None`).
    shuffle : boolean, optional
        Shuffle before splitting into batches or not (default `False`).
    random_state : integer, optional
        Seed variable, only useful if `shuffle = True` (default `None`).

    Returns
    -------
    numpy.ndarray
        Numpy array of shape `(folds, 1(loss) + no_of_metrics)`.

        Contains validation scores evaluated for each fold.
    """

    # Initalize KFold
    kfolds = KFold(n_splits=folds, random_state=random_state, shuffle=shuffle)
    all_metrics = []

    # To build the model
    if type(model).__name__ == 'SVDpp':
        model.implicit_feedback(X[:10, :])
    model(X[:10, :])

    # Workaround to reset weights after each fold fit
    weights = model.get_weights()
    i = 1

    for train, val in kfolds.split(X, y):

        # Gather implicit feedback if model is SVD++
        if type(model).__name__ == 'SVDpp':
            model.implicit_feedback(X[train])

        print(f'\nFitting on Fold {i}')
        # Train and evaluate metrics
        history = model.fit(
            X[train], y[train], batch_size=batch_size, epochs=epochs, callbacks=callbacks)
        print(f'\nEvaluating on Fold {i}')
        fold_score = history.model.evaluate(X[val], y[val])
        all_metrics.append(fold_score)

        # Reset Weights
        model.set_weights(weights)

        i += 1

    all_metrics = np.array(all_metrics)

    for i, metric in enumerate(model.metrics_names):
        print(f'Mean {metric.capitalize()} : {np.mean(all_metrics.T[i])}')

    return all_metrics



import requests
import zipfile
import io
import pandas as pd


def fetch_ml_100k():
    """Fetches the MovieLens 100k Dataset.

    Returns
    -------
    pandas.dataframe
        Of shape `m by 3`, Columns - ``userId``, ``movieId``,``rating``
    """
    r = requests.get(
        'http://files.grouplens.org/datasets/movielens/ml-latest-small.zip')
    z = zipfile.ZipFile(io.BytesIO(r.content))
    z.extractall()

    ratings_csv_path = 'ml-latest-small/ratings.csv'
    fields = ['userId', 'movieId', 'rating']
    df_ratings = pd.read_csv(ratings_csv_path)[fields]

    return df_ratings



from setuptools import setup, find_packages
import pkg_resources

# The below line is used by GH Actions and should not be modified or moved above or below
version = '1.0.0'

readme_path = pkg_resources.resource_filename(__name__, "README.md")
with open(readme_path, encoding="utf8") as f:
    long_description = f.read()


setup(
    name="tfrec",
    version=version,
    description="A recommender library built on top of Tensorflow and \
    Keras with implementations of SVD and SVD++ algorithms.",
    packages=find_packages(),
    url="https://github.com/Praful932/tf-rec",
    author="Praful Mohanan",
    author_email="praful.mohanan@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: MIT License",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.6",
    install_requires=[
        "tensorflow>=2.0",
        "numpy>=1.18.5",
        "pandas>=1.1.5",
        "scikit-learn>=0.22.2",
        "requests",
    ],
)



import unittest
import numpy as np
import tensorflow as tf
from tensorflow import keras
from tfrec.models import SVD, SVDpp
from tfrec.datasets import fetch_ml_100k
from tfrec.utils import preprocess_and_split

seed = 42
tf.random.set_seed(seed)
np.random.seed(seed)
data = fetch_ml_100k()
dataset, user_item_encodings = preprocess_and_split(data, random_state=seed)


class TestModels(unittest.TestCase):
    def setUp(self):
        # Answer to the Ultimate Question of Life, the Universe, and Everything
        self.seed = 42

        tf.random.set_seed(self.seed)
        np.random.seed(self.seed)
        self.data = fetch_ml_100k()
        self.dataset, user_item_encodings = preprocess_and_split(data, random_state=self.seed)

        (self.x_train, self.y_train), (self.x_test, self.y_test) = self.dataset
        self.num_users = len(np.unique(data['userId']))
        self.num_movies = len(np.unique(data['movieId']))
        self.global_mean = np.mean(data['rating'])

        self.batch_size = 128
        self.learning_rate = 0.007
        self.epochs = 1
        self.opt = keras.optimizers.Adam(self.learning_rate)
        self.loss_fn = keras.losses.MeanSquaredError()

    def test_svd(self):
        """Test Custom Model SVD to avoid breaking in future tensorflow versions"""

        model = SVD(self.num_users, self.num_movies, self.global_mean, random_state=self.seed)
        model.compile(loss=self.loss_fn, optimizer=self.opt)
        history = model.fit(self.x_train, self.y_train, batch_size=self.batch_size, epochs=self.epochs)
        # Obvious Test
        self.assertIsInstance(history.model, SVD)

    def test_svdpp(self):
        """Test Custom Model SVD++ to avoid breaking in future tensorflow versions"""

        model = SVDpp(self.num_users, self.num_movies, self.global_mean, random_state=self.seed)
        model.implicit_feedback(self.x_train)
        model.compile(loss=self.loss_fn, optimizer=self.opt)
        history = model.fit(self.x_train, self.y_train, batch_size=self.batch_size, epochs=self.epochs)
        # Obvious Test
        self.assertIsInstance(history.model, SVDpp)
